/* 
   $Id: file.c,v 1.6 2002/12/28 09:18:58 bhepple Exp $

   part of gjots - a gnome/X11 jotter

   Copyright (C) 2002 Robert Hepple 

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <gnome.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <signal.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "main.h"
#include "file.h"
#include "tree.h"

static char		currentFilename[FILENAME_MAX] = "";
static char		errorFileName[64] = "";
static int		pipeError = 0, errorFile = -1;
static char 	*item_marker = "\\NewEntry";
static char 	*folder_marker = "\\NewFolder";
static char 	*folder_end = "\\EndFolder";
static char 	*ccdecrypt = "ccdecrypt";
static char 	*ccencrypt = "ccencrypt";

/********* FILESELECTION *********/

static int fileselection_value;
static char fileSelectionName[FILENAME_MAX];

gboolean
on_fileselection_key_press_event       (GtkWidget       *widget,
                                        GdkEventKey     *event,
                                        gpointer         user_data)
{

	switch (event->keyval) {
	case GDK_y: 		
	case GDK_Y: 		
		fileselection_value = YES;
		return TRUE;
	case GDK_Escape: 		
		fileselection_value = CANCEL;
		return TRUE;
	default: 
		break;
	}

  return FALSE;
}


void
on_fileselectionOkButton_clicked       (GtkButton       *button,
										 gpointer         user_data)
{
	strcpy(fileSelectionName,
		   gtk_file_selection_get_filename(GTK_FILE_SELECTION(gtk_widget_get_toplevel(GTK_WIDGET(button)))));
	fileselection_value = YES;
}

void
on_fileselectionCancelButton_clicked   (GtkButton       *button,
                                        gpointer         user_data)
{
	fileselection_value = CANCEL;
}

int
fileselection(char *defaultFilename, char *title) {
	GtkWidget	    *fileselectionDialog;

	fileselectionDialog = (GtkWidget *)create_fileselection();
	if (title && *title)
		gtk_window_set_title(GTK_WINDOW(fileselectionDialog), title);
	if (defaultFilename && *defaultFilename)
		gtk_file_selection_set_filename(GTK_FILE_SELECTION(fileselectionDialog), 
										defaultFilename);

	fileselection_value = WAITING;
	gtk_widget_show(fileselectionDialog);
	while (fileselection_value == WAITING)
		gtk_main_iteration();
	gtk_widget_destroy(fileselectionDialog);
	while (gtk_events_pending())
		gtk_main_iteration();

	return (fileselection_value == YES);
}

/********* FILESELECTION END *********/

/* read from file and append items after currentItem, unless it has a
 * subtree in which case push them into the subtree (if intoSubtree is
 * set) */
int
readItem(FILE *f, GtkTreeItem *currentItem) {
	char 		buf[MAXBUF];
	GtkTree 	*branch = NULL;
	int			totalRead = 0;

	if (!GTK_IS_TREE_ITEM(currentItem))
		return 0;

	if (!f)
		return 0;

	branch = (GtkTree *)GTK_TREE_ITEM_SUBTREE(currentItem);
	if (!branch)
		branch = (GtkTree *) (GTK_WIDGET(currentItem))->parent;
	if (!branch)
		return 0;

	while (fgets(buf, MAXBUF,f) && !pipeError) {
		int len = strlen(buf);
		totalRead += len;
		if (strncasecmp(buf, item_marker, strlen(item_marker)) == 0) {
			char *title;

			title = buf + strlen(item_marker);
			while (*title && isspace(*title))
				title++;
			currentItem = newItem(branch, currentItem);

			/* use first line as title instead of an argument on \NetEntry line: */
			if (strlen(title) > 1) {
				char *end;

				addTextToPayload(GTK_WIDGET(currentItem), title);
				end = strchr(title, '\n');
				if (end)
					*end = 0;
				tree_item_set_label(currentItem, title);
			}
			continue;
		}
		if (strncasecmp(buf, folder_marker, strlen(folder_marker)) == 0) {
			GtkTreeItem 	*newFolderItem;
			
			newFolderItem = newFolder(currentItem, 0);
			if (readItem(f, newFolderItem))
				return totalRead;
			continue;
		}
		if (strncasecmp(buf, folder_end, strlen(folder_end)) == 0) {
			return 0;
		}
		addTextToPayload(GTK_WIDGET(currentItem), buf);
		if (strlen(tree_item_get_label(currentItem)) == 0) {
			char *p;
			p = strchr(buf, '\n');
			if (p)
				*p = 0;
			tree_item_set_label(currentItem, buf);
		}
	}
	return totalRead;
}

void sighandler(int);

void
sighandler(int sig) {
	pipeError = 1;
	signal(sig, sighandler);
}

static int
check_ccrypt(char *mode) {
	char *path, *sep, *localpath, *bin_name, *path_element;
	char fullpath[FILENAME_MAX];
	struct stat statbuf;

	bin_name = strcmp(mode, "r") == 0? ccdecrypt: ccencrypt;

	if (!(path = getenv("PATH")))
		goto errReturn;

	if (!(localpath = alloca(strlen(path) + 1)))
		goto errReturn;

	strcpy(localpath, path);
	path_element = sep = localpath; 
	while (path_element && *path_element) {
		sep = strchr(path_element, ':');
		if (sep)
			*sep++ = 0;
		strcpy(fullpath, path_element);
		strcat(fullpath, "/");
		strcat(fullpath, bin_name);
		errno = 0;
		if (stat(fullpath, &statbuf) == 0 && (statbuf.st_mode & 01))
			return 0;	
		path_element = sep;
	}

 errReturn:
	sprintf(fullpath, "Can't find %s", bin_name);
	doWarning(fullpath);
	return -1;
}

/* 
 * 
 * fopen_ccrypt: like fopen but filter input/output through ccrypt if
 * the filename matches *.cpt
 * 
 * For encryption, I've used the equivalent of:
 *
 *   {write cleartext} | KEYWORD=password ccencrypt -E KEYWORD >filename.cpt 2>tmpfile
 * 
 * This exposes the password briefly as an environment parameter
 * although it is 'scrubbed' and removed from the gjots process
 * environment immediately after use. According to the ccrypt man page,
 * "on most modern systems" the ps command protects environment
 * parameters from ordinary users. Under Linux, you can 'cat
 * /proc/PID/environ' to see the environment parameters of any process
 * that you own - only root can spy on another's environment so using
 * -E is no less secure than your login password and root's login
 * password.
 * 
 * The only other way to do this (AFAICS) would be:
 * 
 * {write cleartext} >file.gjot 
 * echo password |ccencrypt -k - file.gjot
 * 
 * ... while this provides better protection for the password, this
 * (briefly) exposes the cleartext as a file - as bad as or worse than
 * exposing the key as an environment parameter!
 * 
 * Decrypting is much easier - I use stdin to pass in the password:
 *
 *     echo password |ccdecrypt -c -k - filename.cpt | {read from pipe}
 * 
 * A supplementary channel (eg. stdin=0, stdout=1 stderr=3
 * stdpasswd=3) for ccrypt to read passwords would be a better
 * solution (but ccrypt does not provide it).
 * 	
 */

FILE *
fopen_ccrypt(char *filename, char *mode) {
	char 	*envname = "KEYWORD"; /* should replace this with a random string? */
	char 	password[MAXBUF];
	char 	confirm[MAXBUF];
	char 	*argv[6];
	int		pfd1[2], pfd2[2];
	FILE 	*retval = NULL;
	pid_t 	pid;
	int		reading;
	char	*ext, *feedback;

	errno = 0;
	ext = strrchr(filename, '.');
	if (ext && (strcmp(ext, ".cpt") == 0))
		;
	else
		return(fopen(filename, mode));

	/* So we need a pipe ... */
	if (check_ccrypt(mode)) {
		return NULL;
	}

	reading = strcmp(mode, "w");

	strcpy(errorFileName, "/tmp/gjotsXXXXXX");
	errorFile = mkstemp(errorFileName);
	if (errorFile < 0) {
		doWarning("Can't create temporary file on /tmp");
		return NULL;
	}

	password[0] = 0;
	feedback = NULL;
	do {
		char 	buf[MAXBUF];
		char 	*basename;

		basename = strrchr(filename, '/');
		if (basename)
			basename++;
		else 
			basename = filename;
		sprintf(buf, "Enter password for %s", basename);
		if (OK != doGeneralDialog("gjots: Enter Password", buf, OK|CANCEL, 
								  reading? 1: 2, 1, feedback,
								  "Password: ", password, password, MAXBUF,
								  "Confirm: ",  "",       confirm,  MAXBUF))
			return NULL;
		if (reading || strcmp(password, confirm) == 0)
			break;
		feedback = "Password was not confirmed";
	} while (1);

	if (setenv(envname, password, 1))
		return NULL;

    if (pipe(pfd1) < 0) {
		doWarning("Can't open a pipe.");
		goto cleanup;
	}

	if (reading) {
		if (pipe(pfd2) < 0) {
			doWarning("Can't open the second pipe.");
			goto cleanup;
		}
		argv[0] = ccdecrypt;
		argv[1] = "-c";
		argv[2] = "-k";
		argv[3] = "-";
		argv[4] = filename;
		argv[5] = 0;
	} else {
		argv[0] = ccencrypt;
		argv[1] = "-E";
		argv[2] = envname;
		argv[3] = 0;
	}

	if ((pid = vfork()) == -1) {
		doWarning("Fork failed.");
		goto cleanup;
	}
	if (pid == 0) { 
        /* child - put stdout on pipe, stderr on errorFile */
		close(2);
		dup(errorFile);
		close(errorFile);
		
		if (reading) {
			/* read password from stdin via pipe */
			close(pfd2[1]); /* we don't write to this pipe */
			close(0); /* set stdin to the pipe */
			dup(pfd2[0]);
			close(pfd2[0]);

			/* write cleartext to stdout via pipe */
			close(pfd1[0]); /* we don't read from this pipe */
			close(1); /* set stdout to the pipe */
			dup(pfd1[1]);
			close(pfd1[1]);
		} else {
			int output;

			close(pfd1[1]); /* we don't write to this pipe */
			close(0); /* set stdin to this pipe */
			dup(pfd1[0]);
			close(pfd1[0]);
			/* now set stdout to file: */
			output = open(filename, O_WRONLY|O_CREAT, 0600);
			if (output < 0)
				_exit(errno);
			close(1);	
			dup(output);
			close(output);
		}
		execvp(argv[0], argv);
		_exit(errno); /* abandon hope all ye who enter here */
	} else { /* PARENT */
		sighandler(SIGPIPE); /* mainly so that if there is no ccrypt we don't abort */
		pipeError = 0;

		if (reading) {
			FILE *f;

			/* write password to stdin of ccdecrypt: */
			close(pfd2[0]); /* we don't read from this pipe */
			if ((f = fdopen(pfd2[1], "w")) == NULL) {
				doWarning("Can't write to pipe");
				goto cleanup;
			}
			fputs(password, f); /* it _is_ buffered, after all */
			fclose(f);

			/* read from pipe */
			close(pfd1[1]); /* we don't write to this pipe */
			if ((retval = fdopen(pfd1[0], "r")) == NULL) {
				doWarning("Can't read from pipe");
				goto cleanup;
			}
		} else { /* writing */
			/* write to pipe */
			close(pfd1[0]); /* we don't not read from this pipe */
			if ((retval = fdopen(pfd1[1], "w")) == NULL) {
				doWarning("Can't write to pipe");
				goto cleanup;
			}
		}
	}

 cleanup:
	setenv(envname, "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX", 1);
	unsetenv(envname);
	return retval;
}

static void
make_lockfile_name(char *lockfile, char *filename)
{
	char *p;

	strcpy(lockfile, filename);
	
	/* lockfile = `dirname filename` + ".#" + `basename filename`: */
	p = strrchr(lockfile, '/');
	if (p)
		*++p = 0;
	else
		*lockfile = 0;
	strcat(lockfile, ".#");

	p = strrchr(filename, '/');
	if (p)
		p++;
	else
		p = filename;
	strcat(lockfile, p);
}

/*
 * filename == NULL then use currentFilename
 *
 * Return values:
 * 0 = success
 * 1 = fail
 */
int
unlock_file(char *filename)
{
	char	lockfile[MAXBUF];
	int		fd = -1;
	FILE	*f = NULL;
	pid_t	pid;

	if (filename == NULL)
		filename = currentFilename;

	/* first, make sure we locked this file */
	make_lockfile_name(lockfile, filename);
	fd = open(lockfile, O_RDONLY);
	if (fd < 0)
		return 1;

	if (flock(fd, LOCK_EX) < 0) {
		close(fd);
		return 1;
	}

	f = fdopen(fd, "r");
	if (f == NULL) {
		flock(fd, LOCK_UN);
		close(fd);
		return 1;
	}

	if (fscanf(f, "%d", &pid) == 1) {
		if (pid == getpid()) {
			unlink(lockfile);
			flock(fd, LOCK_UN);
			fclose(f);
			return 0;
		}
	}

	/* fail */
	flock(fd, LOCK_UN);
	fclose(f);
	return 1;
}

/*
 * val = READONLY_CLEAR: clear readonly flag
 * val = READONLY_SET: set readonly flag
 * else: enquire readonly flag
 *
 * Return value: current value of readonly flag
 */
int
readonly(int	val)
{
#define READONLY_STRING " - read only"

	static int 	readonlyState = READONLY_CLEAR;
	GtkWidget	*w;
	gchar		title[MAXBUF], *p;

	switch (val) {

	case READONLY_CLEAR:
		readonlyState = 0;
		gtk_text_set_editable(textDisplay, 1);
		w = lookup_widget(gjots, "readOnly");
		if (w) {
			readonlyCheckButtonGuard = 1;
			gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(w), 0);
			readonlyCheckButtonGuard = 0;
		}
		strcpy(title, GTK_WINDOW(gjots)->title);
		p = strstr(title, READONLY_STRING);
		if (p) {
			*p = 0;
			gtk_window_set_title(GTK_WINDOW(gjots), title);
		}
		break;

	case READONLY_SET:
		if (dirty_data)
			return(readonlyState);
		readonlyState = 1;
		gtk_text_set_editable(textDisplay, 0);
		w = lookup_widget(gjots, "readOnly");
		if (w) {
			readonlyCheckButtonGuard = 1;
			gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(w), 1);
			readonlyCheckButtonGuard = 0;
		}
		strcpy(title, GTK_WINDOW(gjots)->title);
		p = strstr(title, READONLY_STRING);
		if (!p) {
			strcat(title, READONLY_STRING);
			gtk_window_set_title(GTK_WINDOW(gjots), title);
		}
		break;
	}

	return (readonlyState);
}

void
toggle_readonly() {

	while (readonly(READONLY_ENQ)) {
		/* unset the readonly flag */
		pid_t 	reason;
		int		ans;
		char	buf[MAXBUF];

		reason = lock_file(NULL);
		if (reason == 0) { /* success */
			readonly(READONLY_CLEAR);
			return;
		}

		if (reason == 1)
			sprintf(buf, "%s: no permission", currentFilename);
		else
			sprintf(buf, "%s: locked by pid %d", currentFilename, reason);
		ans = doGeneralDialog("gjots: Can't lock file.", 
							  buf,
							  TRYAGAIN|READONLY, 
							  /* num_fields = */ 0, 
							  /* is_password = */ 0, 
							  /* feedback = */ NULL,
							  /* field 1 */ NULL, NULL, NULL, 0,
							  /* field 2 */ NULL, NULL, NULL, 0);
		if (ans == READONLY)
			return;
	}

	/* set the readonly flag */
	if (dirty_data) {
		msg("File has been changed - need to save before going read-only!");
		readonly(READONLY_CLEAR); /* to fix up the check button */
		return;
	}

	unlock_file(NULL);
	readonly(READONLY_SET);
}

/*
 * filename == NULL then use currentFilename
 *
 * return values:
 * 0 = success (possibly by setting readonly)
 * 1 = fail (permission)
 * >1 = fail (already locked)
 */
int
lock_file(char *filename)
{
	char 	lockfile[MAXBUF];
	int		fd = -1;
	FILE	*f = NULL;
	pid_t	pid;
	
	if (filename == NULL)
		filename = currentFilename;

	make_lockfile_name(lockfile, filename);
	fd = open(lockfile, O_RDWR | O_CREAT, 0644);
	if (fd < 0)
		return 1;

	if (flock(fd, LOCK_EX) < 0) {
		close(fd);
		return 1;
	}

	f = fdopen(fd, "r+");
	if (f == NULL) {
		flock(fd, LOCK_UN);
		close(fd);
		return 1;
	}

	if (fscanf(f, "%d", &pid) == 1) {
		if (getpgid(pid) < 0 && errno == ESRCH)
			/* locking process no longer exists */
			goto lockdown;
		flock(fd, LOCK_UN);
		fclose(f);
		return pid;
	}

 lockdown:
	fseek(f, 0L, SEEK_SET);
	fprintf(f, "%d", getpid());
	flock(fd, LOCK_UN);
	fclose(f);
	return 0;
}

int
readFile(char *filename) {
	FILE 		*f;
	char		*basename;
	char 		buf[MAXBUF];

	if (dirty_data) {
		if (!writeFile(0))
			return 0;
	}

	if (filename == NULL || *filename == 0) {
		if (!fileselection(currentFilename, "Open file ..."))
			return -1;

		if (*currentFilename && *fileSelectionName)
			if (strcmp(currentFilename, fileSelectionName) == 0)
				return 0;

		if (*currentFilename && !readonly(READONLY_ENQ))
			unlock_file(currentFilename);

		strcpy(currentFilename, fileSelectionName);
		filename = currentFilename;
	} else {
		if (*currentFilename && !readonly(READONLY_ENQ))
			unlock_file(currentFilename);
	}

	while (!readonly(READONLY_ENQ)) {
		pid_t 	reason;
		int		ans;

		reason = lock_file(filename);
		if (reason == 0) /* success */
			break;
		if (reason == 1)
			sprintf(buf, "%s: no permission", filename);
		else
			sprintf(buf, "%s: locked by pid %d", filename, reason);
		ans = doGeneralDialog("gjots: Can't lock file.", 
							  buf,
							  TRYAGAIN|READONLY, 
							  /* num_fields = */ 0, 
							  /* is_password = */ 0, 
							  /* feedback = */ NULL,
							  /* field 1 */ NULL, NULL, NULL, 0,
							  /* field 2 */ NULL, NULL, NULL, 0);
		if (ans == READONLY)
			readonly(READONLY_SET);
	}

	f = fopen_ccrypt(filename, "r");

	basename = strrchr(filename, '/');
	if (basename)
		basename++;
	else
		basename = filename;

	strcpy(currentFilename, filename);
	if (!f) {
		if (errno == ENOENT) {
			msg("New file");
		} else if (errno) {
			sprintf(buf, "%s: %s", basename, strerror(errno));
			doWarning(buf);
			return -1;
		}
	}

	if (!clearCurrent(0)) {
		if (f)
			fclose(f);
		return -1;
	}

	if (f && readItem(f, root) == 0) {
		if (f)
			fclose(f);
		gtk_window_set_title(GTK_WINDOW(gjots), "gjots");
		/* if (ext)
		   doWarning("Read failed - perhaps ccrypt is not installed?"); */
		return -1;
	} else {
		char *ext;

		ext = strrchr(basename, '.');
		if (ext && (strcmp(ext, ".cpt") == 0))
			*ext = 0;

		tree_item_set_label(root, basename);
		gtk_tree_item_expand(root);
		unselectAll();
		gtk_tree_select_child(GTK_TREE(tree), GTK_WIDGET(root));
		sprintf(buf, "%s: read", basename);
		msg(buf);
		sprintf(buf, "gjots: %s", 
				basename);
		gtk_window_set_title(GTK_WINDOW(gjots), buf);
		readonly(readonly(READONLY_ENQ));
	}
	if (f)
		fclose(f);
	dirty_data = 0;
	return 0;
}

int
doExportFile() {
	char 		exportFilename[FILENAME_MAX];
	FILE		*f;
	char 		*basename;
	GList 		*selection;
	GtkWidget 	*item = NULL;
	int			errorFlag;

	selection = GTK_TREE_SELECTION(tree); 

	if (!selection) {
		msg("nothing selected to export!");
		return 0;
	}

	*exportFilename = 0;
	if (currentFilename) {
		strcpy(exportFilename, currentFilename);
		basename = strrchr(exportFilename, '/');
		if (basename)
			*++basename = 0;
	}
	if (!fileselection(exportFilename, "Export selected items to file ...")) 
		return -1;

	strcpy(exportFilename, fileSelectionName);

	f = fopen(exportFilename, "r");
	if (f) {
		int response;
		char backupFilename[FILENAME_MAX];

		fclose(f);
		response = doConfirm("File already exists. Overwrite?");
		if (response != OK)
			return 0;

		strcpy(backupFilename, exportFilename);
		strcat(backupFilename, "~");
		unlink(backupFilename);
		rename(exportFilename, backupFilename);
	}

	f = fopen_ccrypt(exportFilename, "w");

	basename = strrchr(exportFilename, '/');
	if (basename)
		basename++;
	else
		basename = exportFilename;
	if (!f) {
		if (errno) {
			sprintf(exportFilename, "%s: %s", basename, strerror(errno));
			doWarning(exportFilename);
		}
		return 0;
	}

	errorFlag = 0;
	for (; selection; 
		 selection = selection->next) {
		item = (GtkWidget *)selection->data;
		if (GTK_IS_TREE_ITEM(item))
			if (writeItem(f, GTK_TREE_ITEM(item))) {
				errorFlag = 1;
				break;
			}
	}

	fclose(f);
	if (!errorFlag) {
		sprintf(exportFilename, "%s: exported", basename);
		msg(exportFilename);
	} else {
		doWarning("Export failed!");
	}
	return 1;
}

int
doImportFile() {
	FILE 		*f;
	char		*basename;
	char 		importFilename[FILENAME_MAX];
	GtkTreeItem *item;

	if (readonly(READONLY_ENQ))
		return 0;

	*importFilename = 0;
	if (currentFilename) {
		strcpy(importFilename, currentFilename);
		basename = strrchr(importFilename, '/');
		if (basename)
			*++basename = 0;
	}
	if (!fileselection(importFilename, "Import file ...")) 
		return -1;

	f = fopen_ccrypt(fileSelectionName, "r");

	basename = strrchr(fileSelectionName, '/');
	if (basename)
		basename++;
	else
		basename = fileSelectionName;

	if (!f) {
		if (errno) {
			sprintf(importFilename, "%s: %s", basename, strerror(errno));
			doWarning(importFilename);
		}
		return -1;
	}

	if (!(item = newFolder(NULL, 0))) {
		fclose(f);
		return -1;
	}

	if (readItem(f, item) == 0) {
		doWarning("Import failed!");
	} else {
		gtk_tree_item_expand(item);
		unselectAll();
		gtk_tree_select_child(GTK_TREE(tree), GTK_WIDGET(item));
		sprintf(importFilename, "%s: imported", basename);
		msg(importFilename);
		dirty_data = 1;
	}
	fclose(f);
	return 0;
}

int
writeItem(FILE *f, GtkTreeItem *item) {
	GList 		*children;
	GtkTree		*subTree;
	struct payload_s *payload;

	if (item != root) {
		if (fprintf(f, "%s\n", 
					item_marker) == 0)
			return -1;
	}
	if (pipeError) return -1;

	payload = gtk_object_get_user_data(GTK_OBJECT(item));
	if (payload && payload->text) {
		if (fputs(payload->text, f) == EOF || pipeError)
			return -1;
		if (payload->text[strlen(payload->text) - 1] != '\n')
			if (fputs("\n", f) == EOF || pipeError)
				return -1;
	}

	subTree = (GtkTree *)GTK_TREE_ITEM_SUBTREE(item);
	if (subTree) {
		if (item != root)
			if (fprintf(f, "%s\n", folder_marker) == 0 || pipeError)
				return -1;
		children = gtk_container_children (GTK_CONTAINER(subTree));
		while (children) {
			if (GTK_IS_TREE_ITEM(children->data))
				if (writeItem(f, GTK_TREE_ITEM(children->data)) || pipeError)
					return -1;
			children = g_list_remove_link(children, children);
		}
		if (item != root)
			if (fprintf(f, "%s\n", folder_end) == 0 || pipeError)
				return -1;
	}
	return 0;
}

int
writeFile(int saveAs) {
	char 		buf[MAXBUF];
	FILE		*f;
	char 		*basename;
	int			old_readonly;

	if (*currentFilename == 0) {
		saveAs = 1;
	}

	if (saveAs) {
		if (!fileselection(currentFilename, "Save as ..."))
			return 0;

		f = fopen(fileSelectionName, "r");
		if (f) {
			int response;
			fclose(f);
			response = doConfirm("File already exists. Overwrite?");
			if (response != OK)
				return 0;
		}

		old_readonly = readonly(READONLY_ENQ);

		readonly(READONLY_CLEAR);
		while (!readonly(READONLY_ENQ)) {
			pid_t 	reason;
			int		ans;

			reason = lock_file(fileSelectionName);
			if (reason == 0) /* success */
				break;
			if (reason == 1)
				sprintf(buf, "%s: no permission", fileSelectionName);
			else
				sprintf(buf, "%s: locked by pid %d", fileSelectionName, reason);
			ans = doGeneralDialog("gjots: Can't lock file.", 
								  buf,
								  CANCEL|TRYAGAIN, 
								  /* num_fields = */ 0, 
								  /* is_password = */ 0, 
								  /* feedback = */ NULL,
								  /* field 1 */ NULL, NULL, NULL, 0,
								  /* field 2 */ NULL, NULL, NULL, 0);
			if (ans == CANCEL) {
				readonly(old_readonly);
				return 0;
			}
		}

		if (*currentFilename && !readonly(READONLY_ENQ))
			unlock_file(currentFilename);
		strcpy(currentFilename, fileSelectionName);
	} else if (!dirty_data) {
		msg("No changes to save");
		return 0;
	}

	/* make backup file */
	f = fopen(currentFilename, "r");
	if (f) {
		fclose(f);
		strcpy(buf, currentFilename);
		strcat(buf, "~");
		unlink(buf);
		rename(currentFilename, buf);
	}

	f = fopen_ccrypt(currentFilename, "w");

	basename = strrchr(currentFilename, '/');
	if (basename)
		basename++;
	else
		basename = currentFilename;
	if (!f) {
		if (errno) {
			sprintf(buf, "%s: %s", basename, strerror(errno));
			doWarning(buf);
		}
		return 0;
	}

	if (writeItem(f, root)) {
		doWarning("Save failed!");
	} else {
		sprintf(buf, "%s: saved", basename);
		msg(buf);
		sprintf(buf, "gjots: %s", basename);
		gtk_window_set_title(GTK_WINDOW(gjots), buf);
		readonly(readonly(READONLY_ENQ)); /* fix up the title & check button */
		tree_item_set_label(root, basename);
		dirty_data = 0;
	}
	fclose(f);
	return 1;
}

void
doNewFile() {
	if (readonly(READONLY_ENQ))
		return;

	clearCurrent(1);
	*currentFilename = 0;
}

void
displayFileErrors() {
	char buf[MAXBUF];
	FILE *f;

	struct stat statbuf;

	if (errorFile < 0)
		goto ret;

	errno = 0; /* sometimes set non-zero at this point - by fopen? weird */
	if (stat(errorFileName, &statbuf) < 0)
		goto ret;

	if (statbuf.st_size == 0)
		goto ret;

	if ((f = fopen(errorFileName, "r")) == NULL)
		goto ret;

	if (fgets(buf, MAXBUF, f) == NULL) {
		fclose(f);
		goto ret;
	}
	fclose(f);

	doWarning(buf);

 ret:
	if (errorFileName[0]) {
		unlink(errorFileName);
		errorFileName[0] = 0;
	}
	errorFile = -1;
}
