/* 
   $Id: main.c,v 1.8 2002/12/28 09:18:58 bhepple Exp $

   part of gjots - a gnome/X11 jotter

   Copyright (C) 2002 Robert Hepple 

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

*/

/*
 * Initial main.c file generated by Glade. Edit as required. Glade
 * will not overwrite this file.
 */

/*
 * why? because all the others are either too complex (eg kjots takes
 * sooo long to start up - at least outside KDE), too broken or too
 * simple (eg kjots has single level tree)
 */

/*
 * TODO:
 * cut need confirm? or undo? or paste?
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdio.h>
#include <ctype.h>
#include <signal.h>
#include <gnome.h>
#include <popt.h>

#include "interface.h"
#include "support.h"

#include "main.h"
#include "callbacks.h"
#include "file.h"
#include "tree.h"

GtkWidget 		*gjots = NULL;
GtkTree			*tree = NULL;
GtkTreeItem		*root = NULL;
GtkText			*textDisplay;
GdkFont			*textDisplayFont = NULL;
GtkMenu			*treeItemMenu;
GnomeAppBar 	*appbar = NULL;
static char		rcFilename[FILENAME_MAX];
int				dirty_data = 0;
char 			*newItemDefaultLabel = "New Item";

struct prefs_s prefs = {
	65,					/* lineLength 									*/
	"xterm -e vi %s",	/* external editor 								*/
	"%F",				/* date format 									*/
	"-misc-fixed-medium-r-semicondensed--13-120-75-75-c-60-iso8859-1", /* 'fixed' font */
	/* I prefer -adobe-courier-medium-r-normal-*-*-110-*-*-m-*-iso8859-1*/
	"",					/* findString									*/
	"",					/* replaceString								*/
	0,					/* matchCase									*/
	0,					/* backwardsSearch								*/
	0,					/* findStringIsRegex							*/
	1,					/* findIsGlobal									*/
	211,				/* treeWidth									*/
};

char * 
safe_strncpy(char *dest, char *src, int len) {
	char * retval;

	retval = strncpy(dest, src, len);
	dest[len - 1] = 0;
	return retval;
}

void
read_rc_file(char *prefFile, char *geomString) {
	char 	buf[FILENAME_MAX];
	int		mainWinWidth = -1, mainWinHeight = -1;
	FILE	*f;
	char	*homedir;

	if (prefFile)
		safe_strncpy(rcFilename, prefFile, FILENAME_MAX);
	else {
		homedir = getenv("HOME");
		if (!homedir)
			return;

		sprintf(rcFilename, "%s/%s", homedir, ".gjotsrc");
	}

	f = fopen(rcFilename, "r");
	if (f) {

		while (fgets(buf, FILENAME_MAX, f)) {
			char *b = buf;

			if (*b && buf[strlen(buf) - 1] == '\n')
				buf[strlen(buf) - 1] = 0;

			if (*b == '#')
				continue;

			while (*b && !isspace(*b))
				b++;
			*b++ = 0;
			if (strcmp(buf, "mainWinWidth") == 0) 
				mainWinWidth = atol(b);
			if (strcmp(buf, "mainWinHeight") == 0) 
				mainWinHeight = atol(b);
			if (strcmp(buf, "lineLength") == 0) 
				prefs.lineLength = atol(b);
			if (strcmp(buf, "externalEditor") == 0) 
				safe_strncpy(prefs.externalEditor, b, 1024);
			if (strcmp(buf, "dateFormat") == 0) 
				safe_strncpy(prefs.dateFormat, b, MAXBUF);
			if (strcmp(buf, "fontName") == 0) 
				safe_strncpy(prefs.fontName, b, MAXBUF);

			// if (strcmp(buf, "findString") == 0) 
			// 		safe_strncpy(prefs.findString, b, MAXBUF);
			// if (strcmp(buf, "replaceString") == 0) 
			// 		safe_strncpy(prefs.replaceString, b, MAXBUF);
			if (strcmp(buf, "matchCase") == 0) 
				prefs.matchCase = atol(b);
			if (strcmp(buf, "backwardsSearch") == 0) 
				prefs.backwardsSearch = atol(b);
			if (strcmp(buf, "findStringIsRegex") == 0) 
				prefs.findStringIsRegex = atol(b);
			if (strcmp(buf, "findIsGlobal") == 0) 
				prefs.findIsGlobal = atol(b);

			if (strcmp(buf, "treeWidth") == 0) {
				GtkPaned	*treePane;
				prefs.treeWidth = atol(b);
				treePane = (GtkPaned *)lookup_widget(gjots, "hpaned1");
				if (treePane && prefs.treeWidth > 10)
					gtk_paned_set_position(treePane, prefs.treeWidth);
			}

			if (strcmp(buf, "topToolbar") == 0) {
				GtkWidget	*w;

				w = lookup_widget(gjots, "topToolbarCheck");
				if (w)
					gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(w), atoi(b));
			}

			if (strcmp(buf, "leftToolbar") == 0) {
				GtkWidget	*w;

				w = lookup_widget(gjots, "leftToolbarCheck");
				if (w)
					gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(w), atoi(b));
			}
		}

		fclose(f);
	} else {
		mainWinHeight = 400;
		mainWinWidth = 750;
	}
	
	if (geomString) {
		int x, y;
		printf("geomString=%s\n", geomString);
		gnome_parse_geometry(geomString, &x, &y, &mainWinWidth, &mainWinHeight);
	}

	if (mainWinHeight > 0 && mainWinHeight < 4000 &&
		mainWinWidth > 0 && mainWinWidth < 4000)
		gtk_window_set_default_size(GTK_WINDOW(gjots), mainWinWidth, mainWinHeight);
}

void
write_rc_file() {
	GtkAllocation 	*allocation;
	FILE			*f;

	f = fopen(rcFilename, "w");
	if (!f)
		return;

	allocation = &(gjots->allocation);
		
	fprintf(f, "mainWinWidth %d\n", allocation->width);
	fprintf(f, "mainWinHeight %d\n", allocation->height);
	fprintf(f, "lineLength %d\n", prefs.lineLength);
	fprintf(f, "externalEditor %s\n", prefs.externalEditor);
	fprintf(f, "dateFormat %s\n", prefs.dateFormat);
	fprintf(f, "fontName %s\n", prefs.fontName);
	// fprintf(f, "findString %s\n", prefs.findString);
	// fprintf(f, "replaceString %s\n", prefs.replaceString);
	fprintf(f, "matchCase %d\n", prefs.matchCase);
	fprintf(f, "backwardsSearch %d\n", prefs.backwardsSearch);
	fprintf(f, "findStringIsRegex %d\n", prefs.findStringIsRegex);
	fprintf(f, "findIsGlobal %d\n", prefs.findIsGlobal);

	{
		GtkPaned	*treePane;
		treePane = (GtkPaned *)lookup_widget(gjots, "hpaned1");
		if (treePane && treePane->child1_size > 10)
			prefs.treeWidth = treePane->child1_size;
		fprintf(f, "treeWidth %d\n", prefs.treeWidth);
	}

	{
		GtkWidget	*w;

		w = lookup_widget(gjots, "topToolbarCheck");
		if (w)
			fprintf(f, "topToolbar %d\n", GTK_CHECK_MENU_ITEM(w)->active);
		w = lookup_widget(gjots, "leftToolbarCheck");
		if (w)
			fprintf(f, "leftToolbar %d\n", GTK_CHECK_MENU_ITEM(w)->active);
	}

	fclose(f);
}

void
enablePaste() {
	GtkWidget *w;

	w = lookup_widget(gjots, "paste1");
	if (w)
		gtk_widget_set_sensitive(w, 1);
	w = lookup_widget(gjots, "pasteButton");
	if (w)
		gtk_widget_set_sensitive(w, 1);

#if 0
	{
		int pos;

		w = gnome_app_find_menu_pos(GTK_WIDGET(treeItemMenu), "_Paste", &pos);
		if (w) /* ... no, this is the parent! But how to get the pos'th item? */
			gtk_widget_set_sensitive(w, 1);
	}
#endif
}

void
disablePaste() {
	GtkWidget *w;

	w = lookup_widget(gjots, "paste1");
	if (w)
		gtk_widget_set_sensitive(w, 0);
	w = lookup_widget(gjots, "pasteButton");
	if (w)
		gtk_widget_set_sensitive(w, 0);

#if 0
	{
		int pos;

		w = gnome_app_find_menu_pos(GTK_WIDGET(treeItemMenu), "_Paste", &pos);
		if (w) /* ... no, this is the parent! But how to get the pos'th item? */
			gtk_widget_set_sensitive(w, 0);
	}
#endif
}

static void
sighandler(int sig)
{
	quit_gjots();
}

int
main (int argc, char *argv[]) {
	char 	*geomString = NULL;
	char 	*prefFile = NULL;
	int		main_readonly = 0;

	struct poptOption optionsTable[] = {
		{ "geometry", 'g', POPT_ARG_STRING, &geomString, 0, "Window geometry", "WIDTHxHEIGHT[+X+Y]" },
		{ "prefs", 'p', POPT_ARG_STRING, &prefFile, 0, "Preferences file", "filename" },
		{ "readonly", 'r', POPT_ARG_NONE, &main_readonly, 0, "Read only (no locking)", NULL },
		{ NULL, '\0', 0, NULL, 0, NULL, NULL} /* end the list */
	};
	poptContext optCon;
	char *argFilename;
	gnome_init_with_popt_table("gjots", VERSION, argc, argv,
							   optionsTable, 0, &optCon);

	gjots = create_gjots ();

    /* relies on having a gjots object to resize - must be after create_gjots()*/
	read_rc_file(prefFile, geomString); 

	tree = (GtkTree *)lookup_widget(gjots, "tree");
	textDisplay = (GtkText *)lookup_widget(gjots, "textDisplay");
	gtk_object_set_user_data(GTK_OBJECT(textDisplay), (gpointer *) NULL);
	textDisplayFont = gdk_font_load(prefs.fontName);
	readonly(main_readonly? READONLY_SET: READONLY_CLEAR);
	treeItemMenu = (GtkMenu *)create_treeItemMenu();
	appbar = (GnomeAppBar *)lookup_widget(gjots, "appbar");
	gnome_appbar_set_default(appbar, "gjots Version " VERSION);

	gtk_tree_set_view_mode(GTK_TREE_ROOT_TREE(tree), GTK_TREE_VIEW_ITEM);  
	root = newFolder(NULL, 1);

	/* gtk_tree_item_expand(GTK_TREE_ITEM(root)); */

	argFilename = poptGetArg(optCon);
	if (argFilename) {
		readFile(argFilename);
		displayFileErrors();
	} else {
		char filename[FILENAME_MAX], *homedir;
	  
		if ((homedir = getenv("HOME"))) {
			safe_strncpy(filename, homedir, FILENAME_MAX);
			strcat(filename, "/.gjotsfile");
			readFile(filename);
			displayFileErrors();
		}
	}

	gtk_widget_show (gjots);

	sendFocusToTree();

	signal(SIGINT, sighandler);
	signal(SIGQUIT, sighandler);

	gtk_main ();
	return 0;
}

