/*
   $Id: tree.c,v 1.4 2002/12/28 09:18:58 bhepple Exp $

   part of gjots - a gnome/X11 jotter

   Copyright (C) 2002 Robert Hepple

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

*/

#if 0

GtkTree *tree {
	GtkWidget *tree_owner = NULL
	GList *children --+
}					  |
^  	   	   	   	   	  |
|					  V
|  	   	   	   	      GtkTreeItem *root {  <-------------------------------+
+---<----------------<--- GtkWidget *parent;
                          GtkWidget *subtree --+                           |
                      }                        |                           |
                                               V                           |
                                               GtkTree *subtree1 {         |
                                                   GtkWidget *tree_owner --+
                                                   GList children -------------->
                                               }

#endif

#include <gnome.h>

#include "callbacks.h"
#include "support.h"
#include "main.h"
#include "file.h"
#include "tree.h"

static struct payload_s *
newPayload(char *buf) 
{
	struct payload_s *retval;

	retval = g_malloc(sizeof(struct payload_s));
	retval->text = g_malloc(sizeof(buf) + 1);
	strcpy(retval->text, buf);
	return retval;
}

void
addTextToPayload(GtkWidget *item, char *buf) 
{
	struct payload_s *payload;

	if (!item || !buf || !*buf)
		return;

	payload = (struct payload_s *)gtk_object_get_user_data(GTK_OBJECT(item));

	if (!payload) {
		payload = newPayload(buf);
		gtk_object_set_user_data(GTK_OBJECT(item), (gpointer) payload);
	} else {
		payload->text = g_realloc(payload->text,
								  strlen(payload->text) + strlen(buf) + 1);
		strcat(payload->text, buf);
	}
}

static void
destroyPayload(struct payload_s *payload) 
{
	if (payload) {
		if (payload->text) {
			g_free(payload->text);
			if (readonly(READONLY_ENQ))
				dirty_data = 1;
		}
		free(payload);
	}
}

void
clearTextDisplay() 
{
	gtk_text_freeze(textDisplay);
	gtk_text_set_point(textDisplay, 0);
	gtk_text_forward_delete(textDisplay, gtk_text_get_length(textDisplay));
	gtk_text_thaw(textDisplay);
}

void
syncPayload() 
{
	struct payload_s *payload;
	char *newText;
	GtkWidget *item;

	if (readonly(READONLY_ENQ))
		return;

	newText = gtk_editable_get_chars(GTK_EDITABLE(textDisplay),
									 0,
									 gtk_text_get_length(textDisplay));

	item = (GtkWidget *) gtk_object_get_user_data(GTK_OBJECT(textDisplay));
	if (item) {
		payload = (struct payload_s *)gtk_object_get_user_data(GTK_OBJECT(item));
		if (payload) {
			if (payload->text) {
				if (strcmp(payload->text, newText))
					dirty_data = 1;
				g_free(payload->text);
			}
			payload->text = newText;
		}
	}
}

static int
treeYPosition(GtkWidget *item, int incr) 
{
	int y = 0;
	GtkTree *parent;

	while (item && GTK_IS_TREE_ITEM(item) && (GtkTreeItem *)item != root) {
//		printf("adding %d for <%s>\n", item->allocation.y, tree_item_get_label(item));
		y += item->allocation.y;
		parent = (GtkTree *)item->parent;
		item = (GtkWidget *)parent->tree_owner;
		y += incr;
	}
	return y;
}

/* We probably should be checking the selection is visible on some
 * kind of "tree has moved" signal but what the heck
 */
void
checkOrigin(GtkWidget *item) 
{
	GtkScrolledWindow 	*treeScrolledWindow;
	GtkAdjustment		*vadj;
	int 				y;
	GtkTree 			*parent;
	GtkWidget			*thisItem;

	/* make sure item is expanded */
	thisItem = item;
	while (thisItem != (GtkWidget *) root) {
		parent = (GtkTree*) thisItem->parent;
		thisItem = parent->tree_owner;
		if (GTK_TREE_ITEM(thisItem)->subtree) /* which it must be! */
			gtk_tree_item_expand(GTK_TREE_ITEM(thisItem));
	}


	treeScrolledWindow = (GtkScrolledWindow *)lookup_widget(gjots,
															"treeScrolledWindow");
	if (!treeScrolledWindow)
		return;
	vadj = gtk_scrolled_window_get_vadjustment(treeScrolledWindow);
	if (!vadj)
		return;
#if 0
	printf("lower=%g upper=%g value=%g step=%g page_incr=%g page_size=%g y=%d\n",
		   vadj->lower,
		   vadj->upper,
		   vadj->value,
		   vadj->step_increment,
		   vadj->page_increment,
		   vadj->page_size,
		   item->allocation.y);
#endif
	y = treeYPosition(item, item->allocation.height); /* for root */
	// printf("Total y = %d\n", y);
	if (y < vadj->value) {
		gtk_adjustment_set_value(vadj, y);
		// printf("Is off the top\n");
	}
	if (y + item->allocation.height >
		vadj->value + vadj->page_size) {
		gtk_adjustment_set_value(vadj, y + item->allocation.height - vadj->page_size);
		// printf("off the bottom\n");
	}
}

int
clearCurrent(int autoAddItem) 
{
	if (dirty_data) {
		int status;

		status = writeFile(0);
		displayFileErrors();
		if (status) {
			return 0;
		}
	}
	clearTextDisplay();
	gtk_object_set_user_data(GTK_OBJECT(textDisplay), (gpointer) NULL);

	gtk_tree_remove_item(tree, GTK_WIDGET(root)); /* I assume this frees up everything ... */
	root = NULL;
	root = newFolder(NULL, autoAddItem);
	dirty_data = 0;
	return 1;
}

static GtkWidget *
getChild(GtkTree *t, gint n) 
{
	GList 		*children;
	gint		count = 0;
	GtkWidget	*retval = NULL;

	if (!t || !GTK_IS_TREE(t))
		return NULL;

    children = t->children;
    while (children) {
		if (count == n)
			retval = GTK_WIDGET(children->data);
		children = children->next;
		count++;
	}
	return retval;
}

void
sendFocusToTree() 
{
	GtkWidget 	*item = NULL;

	syncPayload();
	item = (GtkWidget *) gtk_object_get_user_data(GTK_OBJECT(textDisplay));

	if (item) {
		GtkTree *parent = (GtkTree *)item->parent;

		if (!parent)
			return;

		if (item) {
			unselectAll();
			gtk_widget_grab_focus(item);
			gtk_widget_grab_default(item);
			gtk_tree_select_child(parent, item);
			checkOrigin(item);
		}
	}
}

/* bound to Right - 
 * if on a folder 
 *   goto first child 
 * else if there is a next folder
 *  goto next folder
 * else
 *  goto last item in folder
 */
static void
goDownTree() 
{
	GList 		*selection;
	GtkWidget 	*item = NULL;
	GtkWidget	*lastItem = NULL;

	syncPayload();
	selection = getSelectedItems();
	if (!selection || !selection->data)
		return;

	item = GTK_WIDGET(selection->data);

	g_list_free(selection);

	if (GTK_TREE_ITEM_SUBTREE(item)) { /* if it's a folder goto first child */
		item = getChild((GtkTree *)GTK_TREE_ITEM_SUBTREE(item), 0);
	} else { /* not a folder - goto next folder if any */
		GtkTree 	*parent = (GtkTree *)item->parent;
		int 		position = gtk_tree_child_position(parent, item);

		while (item) {
			lastItem = item;
			item = getChild(parent, ++position);
			if (item && GTK_TREE_ITEM_SUBTREE(item)) {
				break;
			}
		}
	}

	if (!item)
		item = lastItem;

	if (item) {
		unselectAll();
		gtk_widget_grab_focus(item);
		gtk_widget_grab_default(item);
		gtk_tree_select_child((GtkTree *)item->parent, item);
		checkOrigin(item);
	}
}

/* bound to Left - goto parent */
static void
goUpTree() 
{
	GList 		*selection;
	GtkWidget 	*item = NULL;
	gint		position;
	GtkTree 	*parent;


	syncPayload();
	selection = getSelectedItems();
	if (!selection || !selection->data)
		return;

	item = GTK_WIDGET(selection->data);

	g_list_free(selection);

	parent = GTK_TREE(item->parent);
	if (!parent)
		return;

	item = parent->tree_owner;
	if (!item)
		return;

	parent = GTK_TREE(item->parent);
	position = gtk_tree_child_position(parent, item);

	item = getChild(parent, position);
	if (item) {
		unselectAll();
		gtk_widget_grab_focus(item);
		gtk_widget_grab_default(item);
		gtk_tree_select_child(parent, item);
		checkOrigin(item);
	}
}

/* Bound to ^Home */
static void
gotoTopOfTree() 
{
	GtkWidget 	*item = (GtkWidget *)root;
	GtkTree 	*parent = tree;

	syncPayload();

	unselectAll();
	gtk_widget_grab_focus(item);
	gtk_widget_grab_default(item);
	gtk_tree_select_child(parent, item);
	checkOrigin(item);
}

/* bound to ^End */
static void
gotoBottomOfTree() 
{
	GtkWidget	*item = NULL;
	GtkWidget	*parent = NULL;
	gint		position;

	syncPayload();
	item = (GtkWidget *) root;
	parent = (GtkWidget *) tree;

	while (1) {
		position = g_list_length(GTK_TREE(parent)->children);
		if (position < 1)
			break;
		item = getChild(GTK_TREE(parent), position - 1); /* last child */
		parent = (GTK_TREE_ITEM(item))->subtree;
		if (!parent)
			break;
		if (! GTK_TREE_ITEM(GTK_TREE(parent)->tree_owner)->expanded)
			break;
	}

	if (item) {
		parent = item->parent;
		unselectAll();
		gtk_widget_grab_focus(item);
		gtk_widget_grab_default(item);
		gtk_tree_select_child(GTK_TREE(parent), item);
		checkOrigin(item);
	}
}

/* bound to ^a, Home */
static void
gotoTopBranch() 
{
	GList 		*selection;
	GtkWidget 	*item = NULL;
	gint		position;
	GtkTree 	*parent;


	syncPayload();
	selection = getSelectedItems();
	if (!selection || !selection->data)
		return;

	item = GTK_WIDGET(selection->data);

	g_list_free(selection);

	parent = (GtkTree *)item->parent;
	if (!parent)
		return;

	position = 0;
	item = getChild(parent, position);
	if (item) {
		unselectAll();
		gtk_widget_grab_focus(item);
		gtk_widget_grab_default(item);
		gtk_tree_select_child(parent, item);
		checkOrigin(item);
	}
}

/* bound to ^e, End */
static void
gotoBottomBranch() 
{
	GList 		*selection;
	GtkWidget 	*item = NULL;
	gint		position;
	GtkTree 	*parent;

	syncPayload();
	selection = getSelectedItems();
	if (!selection || !selection->data)
		return;

	item = GTK_WIDGET(selection->data);

	g_list_free(selection);

	parent = (GtkTree *)item->parent;

	if (!parent)
		return;

	position = g_list_length(parent->children) - 1;
	item = getChild(parent, position);
	if (item) {
		unselectAll();
		gtk_widget_grab_focus(item);
		gtk_widget_grab_default(item);
		gtk_tree_select_child(parent, item);
		checkOrigin(item);
	}
}

/*
 * flag can be SAME_LEVEL ABSOLUTE VISIBLE
 */
GtkWidget *
getPrevItem(GtkWidget *item, int flag) {
	gint		position;

	if (item) {
		GtkTree *parent = (GtkTree *)item->parent;

		if (!parent)
			return NULL;

		position = gtk_tree_child_position(parent, item);
		if (position < 0)
			return NULL;

		if (position == 0) {
			if (flag == SAME_LEVEL)
				return NULL;

			/* first, go up a level if possible */
			if (parent == tree)
				return NULL;
			item = parent->tree_owner;
			if (!item)
				return NULL;
			parent = (GtkTree *)item->parent;
			position = gtk_tree_child_position(parent, item);

		} else {
			position--;

			item = getChild(parent, position);
			if (flag != SAME_LEVEL && GTK_TREE_ITEM_SUBTREE(item)) {
				if (flag == ABSOLUTE || ((GtkTreeItem *)item)->expanded) {
					GtkTreeItem 	*newItem;
					GtkTree 		*newParent;

					/* On moving up to an folder item, go to the last
					 * item in that folder */
					newParent = (GtkTree *) ((GtkTreeItem *)item)->subtree;
					if (newParent != parent && g_list_length(newParent->children) > 0) {
						newItem = (GtkTreeItem *) getChild(newParent, 0);
						if (newItem) {
							item = (GtkWidget *) newItem;
							parent = newParent;
							position = g_list_length(newParent->children) - 1;
						}
					}
				}
			}
		}
		item = getChild(parent, position);
	}

	return(item);
}

/*
 * flag can be SAME_LEVEL ABSOLUTE VISIBLE
 */
GtkWidget *
getNextItem(GtkWidget *item, int flag) 
{
	GtkTree *parent = (GtkTree *)item->parent;
	GList	*children;
	int		position;

	if (!parent)
		return NULL;

	position = gtk_tree_child_position(parent, item);
	if (position < 0)
		return NULL;

	children = parent->children;
	if (children == NULL)
		return NULL;

	/* on moving down from an expanded folder item, go to the
	 * first item in that folder */
	if (flag != SAME_LEVEL &&
		GTK_IS_TREE_ITEM(item) &&
		GTK_TREE_ITEM_SUBTREE(item) &&
		(flag == ABSOLUTE || ((GtkTreeItem *)item)->expanded)) {
		GtkTreeItem 	*newItem;
		GtkTree 		*newParent;

		newParent = (GtkTree *) ((GtkTreeItem *)item)->subtree;
		if (newParent && g_list_length(newParent->children) > 0) {
			newItem = (GtkTreeItem *) getChild(newParent, 0);
			if (newItem) {
				item = (GtkWidget *) newItem;
				parent = newParent;
				position = 0;
			}
		}
	} else if (position >= g_list_length(children) - 1) {
		/* go up a level if possible, then to the next item */
		if (parent == tree)
			return NULL;
		if (flag == SAME_LEVEL)
			return NULL;

		item = parent->tree_owner;
		if (!item)
			return NULL;
		parent = (GtkTree *)item->parent;
		position = gtk_tree_child_position(parent, item) + 1;
	} else
		position++;

	item = getChild(parent, position);
	return item;
}

/* bound to Shift-Down */
static void
extendDown() 
{
	GtkWidget 	*item = NULL;

	syncPayload();

	item = lastSelectedItem();
	if (!item)
		return;

	item = getNextItem(item, SAME_LEVEL);
	if (item) {
		GtkTree *parent = (GtkTree *)item->parent;

		gtk_tree_select_child(parent, item);
		checkOrigin(item);
	}
}

/* bound to Shift-Up */
static void
extendUp() 
{
	GtkWidget 	*item = NULL;

	syncPayload();

	item = firstSelectedItem();
	if (!item)
		return;

	item = getPrevItem(item, SAME_LEVEL);
	if (item) {
		GtkTree *parent = (GtkTree *)item->parent;

		gtk_tree_select_child(parent, item);
		checkOrigin(item);
	}
}

/* bound to ^p, Up */
static void
gotoPrevInTree() 
{
	GtkWidget 	*item = NULL;

	syncPayload();

	item = firstSelectedItem();
	if (!item)
		return;

	item = getPrevItem(item, VISIBLE);
	if (item) {
		GtkTree *parent = (GtkTree *)item->parent;

		unselectAll();
		gtk_widget_grab_focus(item);
		gtk_widget_grab_default(item);
		gtk_tree_select_child(parent, item);
		checkOrigin(item);
	}
}

/* bound to ^n, Down */
static void
gotoNextInTree() 
{
	GtkWidget 	*item = NULL;

	syncPayload();

	item = lastSelectedItem();
	
	if (!item)
		return;

	item = getNextItem(item, VISIBLE);
	if (item) {
		unselectAll();
		gtk_widget_grab_focus(item);
		gtk_widget_grab_default(item);
		gtk_tree_select_child((GtkTree *)item->parent, item);
		checkOrigin(item);
	}
}

/* Only consider items at the same level as the focus item - or if
 * none has the focus then at the same level as the first tree item in
 * the selection. Gets the items in the tree order rather than
 * selected order. Return value needs to be free'd */
GList *
getSelectedItems() {
	GtkWidget 	*item = NULL, *firstItem = NULL;
	GList 		*selection, *iterator, *newList = NULL;
	GtkTree		*parent;

	for (selection = GTK_TREE_SELECTION(tree);
		 selection;
		 selection = selection->next) {
		item = (GtkWidget *)selection->data;
		if (GTK_IS_TREE_ITEM(item)) {
			if (!firstItem)
				firstItem = item;
			if (GTK_WIDGET_HAS_FOCUS(item)) {
				firstItem = item;
				break;
			}
		} else
			item = NULL;
	}		
	if (!firstItem)
		return NULL;
	
	parent = GTK_TREE(firstItem->parent);
	firstItem = NULL;

	for (iterator = parent->children;
		 iterator && iterator->data;
		 iterator = iterator->next) {
		item = GTK_WIDGET(iterator->data);
		
		if (GTK_IS_TREE_ITEM(item) && 
			g_list_find(GTK_TREE_SELECTION(tree), item))
			newList = g_list_append(newList, (gpointer) item);
	}	
	return newList;
}

/* Only consider items at the same level as the focus item - or if
 * none has the focus then at the same level as the first tree item in
 * the selection */
GtkWidget *
firstSelectedItem() {
	GtkWidget 	*firstItem = NULL;
	GList 		*selectedItems;

	selectedItems = getSelectedItems();
	if (!selectedItems)
		return NULL;

	firstItem = selectedItems->data;
	g_list_free(selectedItems);
	return firstItem;
}

/* Only consider items at the same level as the focus item - or if
 * none has the focus then at the same level as the first tree item in
 * the selection */
GtkWidget *
lastSelectedItem() {
	GtkWidget 	*lastItem = NULL;
	GList 		*selectedItems;
	int			last;

	selectedItems = getSelectedItems();
	if (!selectedItems)
		return NULL;

	last = g_list_length(selectedItems);
	if (last > 0)
		lastItem = GTK_WIDGET(g_list_nth(selectedItems, last - 1)->data);
	g_list_free(selectedItems);
	return lastItem;
}

static void
moveItems(GList 	*newSelection, 
		  GtkWidget	*targetItem)
{
	GList 		*iterator;
	FILE		*tmpStorage;
	int			num_selected, focusItem = -1;

	if (!targetItem)
		return;

	tmpStorage = tmpfile();
	if (!tmpStorage) {
		gnome_warning_dialog("Can't open a temporary file");
		return;
	}

	/* remove everything in  newSelection: */
	gtk_widget_hide(GTK_WIDGET(tree));

	for (num_selected = 0, iterator = newSelection;	
		 iterator && iterator->data;
		 iterator = iterator->next, num_selected++) {
		GtkTree		*sourceParent;
		GtkWidget	*item;

		item = GTK_WIDGET(iterator->data);
		if (GTK_WIDGET_HAS_FOCUS(item))
			focusItem = num_selected;

		sourceParent = GTK_TREE(item->parent);

		gtk_tree_unselect_child(sourceParent, item);
		writeItem(tmpStorage, GTK_TREE_ITEM(item));
		gtk_tree_remove_item(sourceParent, item);
		dirty_data = 1;
	}

	/* now re-insert them */
	fseek(tmpStorage, 0L, SEEK_SET);

	if (readItem(tmpStorage, GTK_TREE_ITEM(targetItem)) == 0) {
		gnome_warning_dialog("Move failed!");
		goto cleanup;
	}

	gtk_widget_show(GTK_WIDGET(tree));
	if (focusItem != -1) {
		GtkWidget	*item;
		GtkTree		*parent;
		int			i;

		item = targetItem;
		if (GTK_TREE_ITEM_SUBTREE(item)) {
			parent = GTK_TREE(GTK_TREE_ITEM_SUBTREE(item));
			if (parent && parent->children)
				item = GTK_WIDGET(parent->children->data);
		} else
			item = getNextItem(item, SAME_LEVEL);

		parent = GTK_TREE(item->parent);
		for (i = num_selected; i; i--) {
			gtk_tree_select_child(parent, item);
			if ((i - 1) == focusItem) {
				gtk_widget_grab_focus(item);
				gtk_widget_grab_default(item);
				checkOrigin(item);
			}
			item = getNextItem(item, SAME_LEVEL);
		}
	}

 cleanup:
	fclose(tmpStorage);
}

/* bound to Down (left toolbar), Down (popup menu), ^Down - move items
 * to after next item */
void
moveItemsDown() 
{
	GtkWidget 	*item = NULL;
	GtkWidget	*nextItem;
	int 		numItems, reflate = 0;
	GList		*newSelection = NULL;

	if (readonly(READONLY_ENQ))
		return;

	syncPayload();

	item = lastSelectedItem();
	if (!item || item == GTK_WIDGET(root))
		goto cleanup;

	/* If it has a subtree, collapse it to avoid moving into itself! */
	if (GTK_TREE_ITEM_SUBTREE(item) && GTK_TREE_ITEM(item)->expanded) {
		gtk_tree_item_collapse(GTK_TREE_ITEM(item));
		reflate = 1;
	}

	nextItem = getNextItem(item, VISIBLE);
	if (!nextItem || nextItem == item)
		goto cleanup;

	newSelection = getSelectedItems();
	if (!newSelection)
		goto cleanup;

	numItems = g_list_length(newSelection);
	if (numItems <= 0)
		goto cleanup;

	moveItems(newSelection, nextItem);

	if (reflate) {
		g_list_free(newSelection);
		newSelection = getSelectedItems();
		if (!newSelection)
			goto cleanup;

		item = GTK_WIDGET(newSelection->data);
		if (!item)
			goto cleanup;

		/* If it has a subtree, reflate it */
		if (GTK_TREE_ITEM_SUBTREE(item))
			gtk_tree_item_expand(GTK_TREE_ITEM(item));
	}

 cleanup:
	if (newSelection)
		g_list_free(newSelection);
}

/* bound to "Promote Items" (left toolbar & popup menu) & ^Left - move
 * all selected items up the hierarchy to just after the parent 
 */
void
moveItemsLeft() {
	GtkWidget 	*item = NULL;
	GtkWidget	*dummyItem;
	GList		*newSelection = NULL;
	GtkTree		*targetParent, *grandParent;

	if (readonly(READONLY_ENQ))
		return;

	syncPayload();

	newSelection = getSelectedItems();
	if (!newSelection)
		goto cleanup;

	item = GTK_WIDGET(newSelection->data);
	if (!item || item == GTK_WIDGET(root))
		goto cleanup;

	targetParent = GTK_TREE(item->parent);
	if (!targetParent || targetParent->tree_owner == GTK_WIDGET(root))
		goto cleanup;

	grandParent = (GtkTree *)targetParent->tree_owner->parent;

	/* we need to create a dummy item to stop readItem from dropping
	 * the selection into a directory: */
	dummyItem = gtk_tree_item_new();
	gtk_tree_insert(grandParent, 
					dummyItem, 
					gtk_tree_child_position(grandParent, targetParent->tree_owner) + 1);

	moveItems(newSelection, dummyItem);
	gtk_tree_remove_item(grandParent, dummyItem);

 cleanup:
	if (newSelection)
		g_list_free(newSelection);
}

/* bound to "Demote Items" (left toolbar & popup menu) ^Right - create
 * previous item if none; make previous item a folder if not already;
 * move all selected items into last items of previous item
 */
void
moveItemsRight() {
	GtkWidget 	*item = NULL;
	GtkWidget	*previousItem, *targetItem;
	GtkTree		*targetParent;
	int 		targetPosition;
	GList		*newSelection = NULL;

	if (readonly(READONLY_ENQ))
		return;

	syncPayload();

	newSelection = getSelectedItems();
	if (!newSelection)
		goto cleanup;

	item = GTK_WIDGET(newSelection->data);
	if (!item || item == GTK_WIDGET(root))
		goto cleanup;

	previousItem = getPrevItem(item, SAME_LEVEL);
	if (!previousItem) {
		/* make an item to hold a subtree for the selected items: */
		previousItem = GTK_WIDGET(newItem(GTK_TREE(item->parent), root));
		if (!previousItem)
			goto cleanup;
	}

	/* is it a folder? */
	targetParent = (GtkTree *)GTK_TREE_ITEM_SUBTREE(previousItem);
	if (!targetParent) {
		/* No? so make it one: */
		previousItem = GTK_WIDGET(newFolder(GTK_TREE_ITEM(previousItem), 0));
		if (!previousItem)
			goto cleanup;
		targetParent = GTK_TREE(GTK_TREE_ITEM_SUBTREE(previousItem));
	}
	
	targetPosition = g_list_length(targetParent->children);
	if (targetPosition)
		targetItem = getChild(targetParent, targetPosition - 1);
	else /* is empty */
		targetItem = targetParent->tree_owner;

	moveItems(newSelection, targetItem);

 cleanup:
	if (newSelection)
		g_list_free(newSelection);
}

/* bound to Up (left toolbar), Up (popup menu, ^Up - move items to
 * before previous item */
void
moveItemsUp() 
{
	GtkWidget 	*item = NULL;
	GtkWidget	*previousItem, *targetItem;
	GtkTree		*targetParent;
	int 		targetPosition, numItems, reflate = 0;
	GList		*newSelection = NULL;

	if (readonly(READONLY_ENQ))
		return;

	syncPayload();

	newSelection = getSelectedItems();
	if (!newSelection)
		goto cleanup;

	item = GTK_WIDGET(newSelection->data);
	if (!item || item == GTK_WIDGET(root))
		goto cleanup;

	/* If it has a subtree, collapse it to avoid moving into itself! */
	if (GTK_TREE_ITEM_SUBTREE(item) && GTK_TREE_ITEM(item)->expanded) {
		gtk_tree_item_collapse(GTK_TREE_ITEM(item));
		reflate = 1;
	}

	previousItem = getPrevItem(item, VISIBLE);
	if (!previousItem || previousItem == item)
		goto cleanup;

	numItems = g_list_length(newSelection);
	if (numItems <= 0)
		goto cleanup;

	if (previousItem == GTK_TREE(item->parent)->tree_owner) { 
		/* item was the first in the subtree */
		previousItem = getPrevItem(previousItem, VISIBLE);
		if (!previousItem)
			goto cleanup;
		targetParent = GTK_TREE(previousItem->parent);
		targetPosition = gtk_tree_child_position(targetParent, previousItem) + 1;
	} else {
		/* shift items in same folder: */
		targetParent = GTK_TREE(previousItem->parent);
		targetPosition = gtk_tree_child_position(targetParent, previousItem);
	}

	if (targetPosition)
		targetItem = getChild(targetParent, targetPosition - 1);
	else
		targetItem = targetParent->tree_owner;

	moveItems(newSelection, targetItem);

	if (reflate) {
		g_list_free(newSelection);
		newSelection = getSelectedItems();
		if (!newSelection)
			goto cleanup;

		item = GTK_WIDGET(newSelection->data);
		if (!item)
			goto cleanup;

		/* If it has a subtree, reflate it */
		if (GTK_TREE_ITEM_SUBTREE(item))
			gtk_tree_item_expand(GTK_TREE_ITEM(item));
	}

 cleanup:
	if (newSelection)
		g_list_free(newSelection);
}

/* returns first selected item or NULL */
static GtkWidget *
isInCurrentSelectionTree(GtkWidget *widget) 
{
	GList 		*selection = NULL;
	GtkWidget	*selectedItem;
	GtkTree		*selectionTree, *thisItemTree;

	for (selection = GTK_TREE_SELECTION(tree);
		 selection && !GTK_IS_TREE_ITEM((GtkWidget *)selection->data);
		 selection = selection->next)
		;

	if (GTK_IS_TREE_ITEM(widget) && selection && selection->data &&
		GTK_IS_TREE_ITEM((GtkWidget *)selection->data)) {

		selectedItem = GTK_WIDGET(selection->data);
		selectionTree = (GtkTree *) selectedItem->parent;
		thisItemTree = (GtkTree *) widget->parent;

		if (thisItemTree == selectionTree)
			return (GtkWidget *)selectedItem;
	}
	return NULL;
}

static int
alreadySelected(GtkTree *branch, int pos) 
{
	GList *selection;
	GtkWidget *selectedItem;
	int selectedPos;

	for (selection = GTK_TREE_SELECTION(tree);
		 selection;
		 selection = selection->next) {

		selectedItem = (GtkWidget *)selection->data;
		selectedPos = gtk_tree_child_position(branch, selectedItem);
		if (selectedPos == pos) {
			return 1;
		}
	}
	return 0;
}

/* check the title hasn't changed: */
void
checkTitle(GtkWidget *item)
{
	char 				*end, title[MAXTITLE];
	struct payload_s 	*payload;

	if (!item || !GTK_IS_TREE_ITEM(item) || ((GtkTreeItem *)item) == root)
		return;

	payload = (struct payload_s *)gtk_object_get_user_data(GTK_OBJECT(item));
	if (payload && payload->text) {
		strncpy(title, payload->text, MAXTITLE);
		end = strchr(title, '\n');
		if (end)
			*end = 0;
		else
			title[MAXTITLE - 1] = 0;
		if (strcmp(title, tree_item_get_label(GTK_TREE_ITEM(item)))) {
			tree_item_set_label(GTK_TREE_ITEM(item), title);
		}
	}
}

static void
expandCollapse(GtkWidget *w) 
{
	GtkTreeItem *t;

	if (GTK_IS_TREE(w))
		w = ((GtkTree *)w)->tree_owner;

	if (!GTK_IS_TREE_ITEM(w))
		return;

	t = (GtkTreeItem *) w;
	if (t->expanded)
		gtk_tree_item_collapse(t);
	else
		gtk_tree_item_expand(t);
}

/* call this _after_ the default handler to defeat it Note that most
 * keys cause the textDisplay to get focus - so return TRUE to prevent
 * further processing.
 */
static gint
on_tree_item_key_pressed(GtkWidget *widget, GdkEventKey *event, gpointer user_data) 
{

	unmsg();
	if (GDK_CONTROL_MASK & event->state) {
		switch (event->keyval) {
		case GDK_Left:
		case GDK_KP_Left:
			moveItemsLeft();
			return TRUE;
		case GDK_Right:
		case GDK_KP_Right:
			moveItemsRight();
			return TRUE;
		case GDK_Down:
		case GDK_KP_Down:
			moveItemsDown();
			return TRUE;
		case GDK_Up:
		case GDK_KP_Up:
			moveItemsUp();
			return TRUE;
		case GDK_Home:
		case GDK_KP_Home:
			gotoTopOfTree();
			return TRUE;
		case GDK_End:
		case GDK_KP_End:
			gotoBottomOfTree();
			return TRUE;
		case GDK_A:
		case GDK_a:
			gotoTopBranch();
			return TRUE;
		case GDK_C:
		case GDK_c:
			doCopy();
			return TRUE;
		case GDK_E:
		case GDK_e:
			gotoBottomBranch();
			return TRUE;
		case GDK_N:
		case GDK_n:
			gotoNextInTree();
			return TRUE;
		case GDK_P:
		case GDK_p:
			gotoPrevInTree();
			return TRUE;
		case GDK_V:
		case GDK_v:
			doPaste();
			return TRUE;
		case GDK_X:
		case GDK_x:
			doCut();
			return TRUE;
		default:
			break;
		}

	} else if (GDK_SHIFT_MASK & event->state) {
		switch (event->keyval) {
		case GDK_Down:
		case GDK_KP_Down:
			extendDown();
			return TRUE;
		case GDK_Up:
		case GDK_KP_Up:
			extendUp();
			return TRUE;
		}
	} else if ((GDK_MODIFIER_MASK & event->state) == 0) { /* plain key */
		switch (event->keyval) {
		case GDK_Down:
		case GDK_KP_Down:
			gotoNextInTree();
			return TRUE;
		case GDK_Up:
		case GDK_KP_Up:
			gotoPrevInTree();
			return TRUE;
		case GDK_Home:
		case GDK_KP_Home:
			gotoTopBranch();
			return TRUE;
		case GDK_End:
		case GDK_KP_End:
			gotoBottomBranch();
			return TRUE;
		case GDK_Left:
		case GDK_KP_Left:
			goUpTree();
			return TRUE;
		case GDK_Right:
		case GDK_KP_Right:
			goDownTree();
			return TRUE;
		case GDK_KP_Enter:
		case GDK_Return:
			expandCollapse(widget);
			return TRUE;
		case GDK_Delete:
		case GDK_KP_Delete:
			doCut();
			return TRUE;
		case GDK_Insert:
		case GDK_KP_Insert:
			doPaste();
			return TRUE;
		case GDK_F12:
			gtk_widget_grab_focus((GtkWidget *)textDisplay);
			return TRUE;
		default:
			break;
		}
	}
	gtk_widget_grab_focus(widget);
	gtk_widget_grab_default(widget);
	return FALSE; /* otherwise global keys (eg ^q) are not seen */
}

static void
safeSelect(GtkWidget *widget) 
{
	GList *selection;
	GtkWidget *selectedItem, *branch;
	int isSelected = 0;

	for (selection = GTK_TREE_SELECTION(tree);
		 selection;
		 selection = selection->next) {

		selectedItem = (GtkWidget *)selection->data;
		if (selectedItem == widget) {
			isSelected = 1;
			break;
		}
	}

	if (!isSelected) {
		branch = widget->parent;
		if (GTK_IS_TREE(branch) && GTK_IS_TREE_ITEM(widget))
			gtk_tree_select_child(GTK_TREE(branch), widget);
	}
}

static gint
on_tree_item_button_pressed(GtkWidget *widget, GdkEvent *event) 
{
	GdkEventButton *event_button;

	unmsg();
	g_return_val_if_fail (widget != NULL, FALSE);
	g_return_val_if_fail (GTK_IS_TREE_ITEM (widget), FALSE);
	g_return_val_if_fail (event != NULL, FALSE);

	if (event->type == GDK_BUTTON_PRESS) {
		event_button = (GdkEventButton *) event;
		switch (event_button->button) {
		case 1:
			if (GDK_CONTROL_MASK & event_button->state) {
				/* Add this item to the selection if in the same tree */
				if (!isInCurrentSelectionTree(widget)) {
					unselectAll();
				}

			} else if (GDK_SHIFT_MASK & event_button->state) {
				/* select all items between the selection & this item
				 * provided they're in the same tree */
				GtkWidget	*selectedItem;
				GtkTree		*thisItemTree = (GtkTree *)widget->parent;

				if ((selectedItem = isInCurrentSelectionTree(widget)) &&
					selectedItem != widget &&
					thisItemTree) {
					int start, end, i, inc;
					start = gtk_tree_child_position(thisItemTree, selectedItem);
					end = gtk_tree_child_position(thisItemTree, widget);
					inc = (start > end)? -1: 1;
					for (i = start ; i != end; i += inc) {
						if (!alreadySelected(thisItemTree, i))
							gtk_tree_select_item(thisItemTree, i);
					}
				}

			} else /* plain button - just select one item */
				unselectAll();
			break;

		case 3:
			if (GDK_CONTROL_MASK & event_button->state)
				;
			else if (GDK_SHIFT_MASK & event_button->state)
				;
			else
				;
			safeSelect(widget);
			gtk_menu_popup(treeItemMenu, NULL, NULL, NULL, NULL,
						   event_button->button, event_button->time);
			return TRUE;
		}
    }
	return FALSE;
}

static void
on_tree_item_destroy(GtkWidget *item,
			gchar     *signame ) 
{
	struct payload_s *payload;
	GtkWidget		 *currentItem = NULL;

	currentItem = (GtkWidget *) gtk_object_get_user_data(GTK_OBJECT(textDisplay));

	if (currentItem == item) {
		clearTextDisplay();
		gtk_object_set_user_data(GTK_OBJECT(textDisplay), (gpointer) NULL);
	}

	payload = (struct payload_s *)gtk_object_get_user_data(GTK_OBJECT(item));
	destroyPayload(payload);
}

static void
on_tree_item_select(GtkWidget *item, gchar *signame ) 
{
	struct payload_s *payload;

	unmsg();
	if (GTK_IS_TREE_ITEM(item)) {
		payload = (struct payload_s *)gtk_object_get_user_data(GTK_OBJECT(item));
		if (payload) {
			syncPayload();
			clearTextDisplay();
			gtk_text_freeze(textDisplay);
			gtk_text_insert(textDisplay, textDisplayFont, NULL, NULL, payload->text, -1);
			gtk_text_thaw(textDisplay);
			gtk_object_set_user_data(GTK_OBJECT(textDisplay), (gpointer) item);
		}
	}
}

/*
 * branch = NULL means use selection to find where to put it
 * if selection is an open folder, append objects to folder
 * if selection is an item, append objects after item
 * else put into 'branch'
 * Place item after afterItem: set afterItem = root to place at 
 * first position.
 */
GtkTreeItem *
newItem(GtkTree *branch, GtkTreeItem *afterItem) 
{
	GList 		*selection = NULL;
	GtkWidget 	*newItem = NULL;
	int			position = 0;

	syncPayload();
	if (branch == NULL) { /* use selection to decide where to put it */
		if (root) {
			selection = GTK_TREE_SELECTION(tree);
			branch = (GtkTree *)GTK_TREE_ITEM_SUBTREE(root);
			if (!branch) {
				/* very first item - create first subtree */
				branch = (GtkTree *) gtk_tree_new();
				if (!branch)
					return NULL;
				gtk_tree_item_set_subtree(root, GTK_WIDGET(branch));
				selection = NULL;
			}
		} else
			branch = tree; /* when creating root */

		if (selection) {
			GtkWidget *item;

			item = (GtkWidget *)selection->data;
			if (item && GTK_IS_TREE_ITEM(item)) {
				branch = (GtkTree *)GTK_TREE_ITEM_SUBTREE(item);
				if (!branch || !GTK_TREE_ITEM(item)->expanded) { 
					/* item is not a folder or is closed */
					branch = (GtkTree *)item->parent;
					if (branch)
						position = gtk_tree_child_position(branch, item) + 1;
				}
			}
		}
	} else {
        /* adding to an existing branch - if afterItem==NULL then at the end */
		if (afterItem) {
			position = gtk_tree_child_position(branch, GTK_WIDGET(afterItem)) + 1;
		} else
			position = g_list_length(branch->children);
	}

	newItem = gtk_tree_item_new_with_label("");
	if (newItem) {
		struct payload_s *payload;

		payload = newPayload("");
		gtk_object_set_user_data(GTK_OBJECT(newItem), (gpointer) payload);

		/* Set the following to get the treeitems to accept key
		 * presses:
		 */
		GTK_WIDGET_SET_FLAGS(newItem, GTK_CAN_DEFAULT);
		GTK_WIDGET_SET_FLAGS(newItem, GTK_CAN_FOCUS);
		gtk_widget_grab_focus (newItem);
		gtk_widget_grab_default (newItem);
		gtk_signal_connect_after(GTK_OBJECT(newItem), "key_press_event",
								 GTK_SIGNAL_FUNC(on_tree_item_key_pressed), "key_press_event");
		gtk_signal_connect (GTK_OBJECT(newItem), "select",
							GTK_SIGNAL_FUNC(on_tree_item_select), "select");
		gtk_signal_connect (GTK_OBJECT(newItem), "destroy",
							GTK_SIGNAL_FUNC(on_tree_item_destroy), "destroy");
		gtk_signal_connect(GTK_OBJECT(newItem), "button_press_event",
						   GTK_SIGNAL_FUNC(on_tree_item_button_pressed), "button_press_event");
		gtk_tree_insert(GTK_TREE(branch), newItem, position);
		gtk_widget_show(newItem);
	}

	return (GtkTreeItem *)newItem;
}

/* bound to New Page (left toolbar), New Page (popup menu) */
void
doNewItem() 
{
	GtkTreeItem *item;

	if (readonly(READONLY_ENQ))
		return;

	item = newItem(NULL, NULL);
	unselectAll();
	gtk_tree_select_child(tree, GTK_WIDGET(item));
	dirty_data = 1;
}

/*
 * if leafItem = NULL use the selection to decide where to put it.
 */
GtkTreeItem *
newFolder(GtkTreeItem *leafItem, int autoAddItem) 
{
	GtkTree 		*branch;

	branch = (GtkTree *) gtk_tree_new();
	if (!branch)
		return NULL;

	if (leafItem == NULL)
		leafItem = newItem(NULL, leafItem);

	if (leafItem) {
		gtk_tree_item_set_subtree(leafItem, GTK_WIDGET(branch));
		if (autoAddItem) {
			newItem(branch, NULL);
		}
		/* gtk_tree_item_expand(leafItem); */
	}

	return leafItem;
}

/* bound to New Folder (left toolbar), New Folder (popup menu) */
void
doNewFolder() 
{
	GtkTreeItem *item;

	if (readonly(READONLY_ENQ))
		return;

	item = newFolder(NULL, 1);
	unselectAll();
	gtk_tree_select_child(tree, GTK_WIDGET(item));
	dirty_data = 1;
}

char *
tree_item_get_label(GtkTreeItem *item) 
{
	GtkLabel *label;
	char *retval = NULL;

	label = GTK_LABEL (GTK_BIN (item)->child);
	if (label)
		gtk_label_get(label, &retval);
	return retval;
}

void
tree_item_set_label(GtkTreeItem *item, char *newLabel) 
{
	GtkLabel *label;

	label = GTK_LABEL (GTK_BIN (item)->child);
	if (label)
		gtk_label_set_text(label, newLabel);
}

void
unselectAll() 
{
	GList 		*selection;
	GtkWidget 	*item = NULL;

	for (selection = GTK_TREE_SELECTION(tree);
		 selection;
		 selection = GTK_TREE_SELECTION(tree)) {

		item = (GtkWidget *)selection->data;
//		if (GTK_IS_TREE_ITEM(item) && GTK_IS_TREE(item->parent))
			gtk_tree_unselect_child(tree, item);
//		else
//			selection = selection->next;
	}
}
